/*
 * kgr_patch_cve-2014-9710
 *
 * Fix for CVE-2014-9710, bsc#939260
 *
 *  Copyright (c) 2015 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <asm/unaligned.h>

#include "ctree.h"
#include "locking.h"
#include "transaction.h"

static void (*kgr_btrfs_tree_read_lock)(struct extent_buffer *eb);
static struct extent_buffer *(*kgr_btrfs_read_lock_root_node)(struct btrfs_root *root);
static void (*kgr_btrfs_tree_read_unlock)(struct extent_buffer *eb);
static int (*kgr_bin_search)(struct extent_buffer *eb, struct btrfs_key *key,
		int level, int *slot);
static void (*kgr_key_search_validate)(struct extent_buffer *b,
		struct btrfs_key *key, int level);
static void (*kgr_reada_for_balance)(struct btrfs_root *root,
		struct btrfs_path *path, int level);
static int (*kgr_split_node)(struct btrfs_trans_handle *trans,
		struct btrfs_root *root, struct btrfs_path *path, int level);
static int (*kgr_balance_level)(struct btrfs_trans_handle *trans,
		struct btrfs_root *root, struct btrfs_path *path, int level);
static void (*kgr_unlock_up)(struct btrfs_path *path, int level,
		int lowest_unlock, int min_write_lock_level, int *write_lock_level);
static struct extent_buffer *(*kgr_btrfs_find_tree_block)(struct btrfs_root *root,
		u64 bytenr, u32 blocksize);
static int (*kgr_btrfs_buffer_uptodate)(struct extent_buffer *buf, u64 parent_transid,
		int atomic);
static int (*kgr_btrfs_read_buffer)(struct extent_buffer *buf, u64 parent_transid);
static int (*kgr_readahead_tree_block)(struct btrfs_root *root, u64 bytenr, u32 blocksize,
		u64 parent_transid);
static struct extent_buffer *(*kgr_read_tree_block)(struct btrfs_root *root, u64 bytenr,
		u32 blocksize, u64 parent_transid);
static int (*kgr_split_leaf)(struct btrfs_trans_handle *trans,
		struct btrfs_root *root, struct btrfs_key *ins_key,
		struct btrfs_path *path, int data_size, int extend);
static void (*kgr_btrfs_tree_lock)(struct extent_buffer *eb);
static void (*kgr_btrfs_tree_unlock)(struct extent_buffer *eb);
static struct extent_buffer *(*kgr_btrfs_lock_root_node)(struct btrfs_root *root);
static int (*kgr_btrfs_leaf_free_space)(struct btrfs_root *root,
		struct extent_buffer *leaf);
static int (*kgr_btrfs_try_tree_write_lock)(struct extent_buffer *eb);
static void (*kgr_btrfs_clear_path_blocking)(struct btrfs_path *p,
		struct extent_buffer *held, int held_rw);
static int (*kgr_btrfs_cow_block)(struct btrfs_trans_handle *trans,
		struct btrfs_root *root, struct extent_buffer *buf,
		struct extent_buffer *parent, int parent_slot,
		struct extent_buffer **cow_ret);
static struct extent_buffer *(*kgr_btrfs_root_node)(struct btrfs_root *root);
static void (*kgr_btrfs_tree_read_unlock_blocking)(struct extent_buffer *eb);
static int (*kgr_btrfs_try_tree_read_lock)(struct extent_buffer *eb);
static void (*kgr_btrfs_node_key)(struct extent_buffer *eb,
		struct btrfs_disk_key *disk_key, int nr);
static void (*kgr_btrfs_set_path_blocking)(struct btrfs_path *p);
static void (*kgr_btrfs_release_path)(struct btrfs_path *p);
static void (*kgr_btrfs_unlock_up_safe)(struct btrfs_path *path, int level);
static void (*kgr_free_extent_buffer)(struct extent_buffer *eb);
static void (*kgr_read_extent_buffer)(struct extent_buffer *eb, void *dstv,
		unsigned long start, unsigned long len);
static int (*kgr_map_private_extent_buffer)(struct extent_buffer *eb, unsigned long start,
		unsigned long min_len, char **map, unsigned long *map_start,
		unsigned long *map_len);
static void (*kgr_btrfs_mark_buffer_dirty)(struct extent_buffer *buf);
static int (*kgr_verify_dir_item)(struct btrfs_root *root,
		struct extent_buffer *leaf, struct btrfs_dir_item *dir_item);
static int (*kgr_btrfs_delete_one_dir_name)(struct btrfs_trans_handle *trans,
		struct btrfs_root *root, struct btrfs_path *path,
		struct btrfs_dir_item *di);
static void (*kgr_btrfs_extend_item)(struct btrfs_root *root,
		struct btrfs_path *path, u32 data_size);
static void (*kgr_btrfs_truncate_item)(struct btrfs_root *root, struct btrfs_path *path,
		u32 new_size, int from_end);
static void (*kgr_write_extent_buffer)(struct extent_buffer *eb, const void *srcv,
		unsigned long start, unsigned long len);
static void (*kgr_btrfs_free_path)(struct btrfs_path *p);
static int (*kgr_btrfs_insert_xattr_item)(struct btrfs_trans_handle *trans,
		struct btrfs_root *root, struct btrfs_path *path, u64 objectid,
		const char *name, u16 name_len, const void *data, u16 data_len);
static void (*kgr_btrfs_assert_tree_locked)(struct extent_buffer *eb);
static struct btrfs_dir_item *(*kgr_btrfs_lookup_xattr)(struct btrfs_trans_handle *trans,
		struct btrfs_root *root, struct btrfs_path *path, u64 dir,
		const char *name, u16 name_len, int mod);
static struct btrfs_path *(*kgr_btrfs_alloc_path)(void);
static int (*kgr_memcmp_extent_buffer)(struct extent_buffer *eb, const void *ptrv,
		unsigned long start, unsigned long len);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "btrfs_tree_read_lock", (void *)&kgr_btrfs_tree_read_lock },
	{ "btrfs_read_lock_root_node", (void *)&kgr_btrfs_read_lock_root_node },
	{ "btrfs_tree_read_unlock", (void *)&kgr_btrfs_tree_read_unlock },
	{ "btrfs:bin_search", (void *)&kgr_bin_search },
	{ "btrfs:key_search_validate", (void *)&kgr_key_search_validate },
	{ "btrfs:reada_for_balance", (void *)&kgr_reada_for_balance },
	{ "btrfs:split_node", (void *)&kgr_split_node },
	{ "btrfs:balance_level", (void *)&kgr_balance_level },
	{ "btrfs:unlock_up", (void *)&kgr_unlock_up },
	{ "btrfs_find_tree_block", (void *)&kgr_btrfs_find_tree_block },
	{ "btrfs_buffer_uptodate", (void *)&kgr_btrfs_buffer_uptodate },
	{ "btrfs_read_buffer", (void *)&kgr_btrfs_read_buffer },
	{ "btrfs:readahead_tree_block", (void *)&kgr_readahead_tree_block },
	{ "btrfs:read_tree_block", (void *)&kgr_read_tree_block },
	{ "btrfs:split_leaf", (void *)&kgr_split_leaf },
	{ "btrfs_tree_lock", (void *)&kgr_btrfs_tree_lock },
	{ "btrfs_tree_unlock", (void *)&kgr_btrfs_tree_unlock },
	{ "btrfs_lock_root_node", (void *)&kgr_btrfs_lock_root_node },
	{ "btrfs_leaf_free_space", (void *)&kgr_btrfs_leaf_free_space },
	{ "btrfs_try_tree_write_lock", (void *)&kgr_btrfs_try_tree_write_lock },
	{ "btrfs_clear_path_blocking", (void *)&kgr_btrfs_clear_path_blocking },
	{ "btrfs_cow_block", (void *)&kgr_btrfs_cow_block },
	{ "btrfs_root_node", (void *)&kgr_btrfs_root_node },
	{ "btrfs_tree_read_unlock_blocking", (void *)&kgr_btrfs_tree_read_unlock_blocking },
	{ "btrfs_try_tree_read_lock", (void *)&kgr_btrfs_try_tree_read_lock },
	{ "btrfs_node_key", (void *)&kgr_btrfs_node_key },
	{ "btrfs_set_path_blocking", (void *)&kgr_btrfs_set_path_blocking },
	{ "btrfs_release_path", (void *)&kgr_btrfs_release_path },
	{ "btrfs_unlock_up_safe", (void *)&kgr_btrfs_unlock_up_safe },
	{ "btrfs:free_extent_buffer", (void *)&kgr_free_extent_buffer },
	{ "btrfs:read_extent_buffer", (void *)&kgr_read_extent_buffer },
	{ "btrfs:map_private_extent_buffer", (void *)&kgr_map_private_extent_buffer },
	{ "btrfs_mark_buffer_dirty", (void *)&kgr_btrfs_mark_buffer_dirty },
	{ "btrfs:verify_dir_item", (void *)&kgr_verify_dir_item },
	{ "btrfs_delete_one_dir_name", (void *)&kgr_btrfs_delete_one_dir_name },
	{ "btrfs_extend_item", (void *)&kgr_btrfs_extend_item },
	{ "btrfs_truncate_item", (void *)&kgr_btrfs_truncate_item },
	{ "btrfs:write_extent_buffer", (void *)&kgr_write_extent_buffer },
	{ "btrfs_free_path", (void *)&kgr_btrfs_free_path },
	{ "btrfs_insert_xattr_item", (void *)&kgr_btrfs_insert_xattr_item },
	{ "btrfs_assert_tree_locked", (void *)&kgr_btrfs_assert_tree_locked },
	{ "btrfs_lookup_xattr", (void *)&kgr_btrfs_lookup_xattr },
	{ "btrfs_alloc_path", (void *)&kgr_btrfs_alloc_path },
	{ "btrfs:memcmp_extent_buffer", (void *)&kgr_memcmp_extent_buffer },
};

#define DEFINE_BTRFS_SETGET_BITS(bits)					\
u##bits btrfs_get_token_##bits(struct extent_buffer *eb, void *ptr,	\
			       unsigned long off,			\
			       struct btrfs_map_token *token)		\
{									\
	unsigned long part_offset = (unsigned long)ptr;			\
	unsigned long offset = part_offset + off;			\
	void *p;							\
	int err;							\
	char *kaddr;							\
	unsigned long map_start;					\
	unsigned long map_len;						\
	int size = sizeof(u##bits);					\
	u##bits res;							\
									\
	if (token && token->kaddr && token->offset <= offset &&		\
	    token->eb == eb &&						\
	   (token->offset + PAGE_CACHE_SIZE >= offset + size)) {	\
		kaddr = token->kaddr;					\
		p = kaddr + part_offset - token->offset;		\
		res = get_unaligned_le##bits(p + off);			\
		return res;						\
	}								\
	err = kgr_map_private_extent_buffer(eb, offset, size,		\
					&kaddr, &map_start, &map_len);	\
	if (err) {							\
		__le##bits leres;					\
									\
		kgr_read_extent_buffer(eb, &leres, offset, size);	\
		return le##bits##_to_cpu(leres);			\
	}								\
	p = kaddr + part_offset - map_start;				\
	res = get_unaligned_le##bits(p + off);				\
	if (token) {							\
		token->kaddr = kaddr;					\
		token->offset = map_start;				\
		token->eb = eb;						\
	}								\
	return res;							\
}									\
void btrfs_set_token_##bits(struct extent_buffer *eb,			\
			    void *ptr, unsigned long off, u##bits val,	\
			    struct btrfs_map_token *token)		\
{									\
	unsigned long part_offset = (unsigned long)ptr;			\
	unsigned long offset = part_offset + off;			\
	void *p;							\
	int err;							\
	char *kaddr;							\
	unsigned long map_start;					\
	unsigned long map_len;						\
	int size = sizeof(u##bits);					\
									\
	if (token && token->kaddr && token->offset <= offset &&		\
	    token->eb == eb &&						\
	   (token->offset + PAGE_CACHE_SIZE >= offset + size)) {	\
		kaddr = token->kaddr;					\
		p = kaddr + part_offset - token->offset;		\
		put_unaligned_le##bits(val, p + off);			\
		return;							\
	}								\
	err = kgr_map_private_extent_buffer(eb, offset, size,		\
			&kaddr, &map_start, &map_len);			\
	if (err) {							\
		__le##bits val2;					\
									\
		val2 = cpu_to_le##bits(val);				\
		kgr_write_extent_buffer(eb, &val2, offset, size);	\
		return;							\
	}								\
	p = kaddr + part_offset - map_start;				\
	put_unaligned_le##bits(val, p + off);				\
	if (token) {							\
		token->kaddr = kaddr;					\
		token->offset = map_start;				\
		token->eb = eb;						\
	}								\
}

DEFINE_BTRFS_SETGET_BITS(16)
DEFINE_BTRFS_SETGET_BITS(32)
DEFINE_BTRFS_SETGET_BITS(64)

static inline int kgr_should_cow_block(struct btrfs_trans_handle *trans,
				   struct btrfs_root *root,
				   struct extent_buffer *buf)
{
#ifdef CONFIG_BTRFS_FS_RUN_SANITY_TESTS
	if (unlikely(test_bit(BTRFS_ROOT_DUMMY_ROOT, &root->state)))
		return 0;
#endif
	/* ensure we can see the force_cow */
	smp_rmb();

	/*
	 * We do not need to cow a block if
	 * 1) this block is not created or changed in this transaction;
	 * 2) this block does not belong to TREE_RELOC tree;
	 * 3) the root is not forced COW.
	 *
	 * What is forced COW:
	 *    when we create snapshot during commiting the transaction,
	 *    after we've finished coping src root, we must COW the shared
	 *    block to ensure the metadata consistency.
	 */
	if (btrfs_header_generation(buf) == trans->transid &&
	    !btrfs_header_flag(buf, BTRFS_HEADER_FLAG_WRITTEN) &&
	    !(root->root_key.objectid != BTRFS_TREE_RELOC_OBJECTID &&
	      btrfs_header_flag(buf, BTRFS_HEADER_FLAG_RELOC)) &&
	    !test_bit(BTRFS_ROOT_FORCE_COW, &root->state))
		return 0;
	return 1;
}

static int kgr_key_search(struct extent_buffer *b, struct btrfs_key *key,
		      int level, int *prev_cmp, int *slot)
{
	if (*prev_cmp != 0) {
		*prev_cmp = kgr_bin_search(b, key, level, slot);
		return *prev_cmp;
	}

	kgr_key_search_validate(b, key, level);
	*slot = 0;

	return 0;
}

static int
kgr_setup_nodes_for_search(struct btrfs_trans_handle *trans,
		       struct btrfs_root *root, struct btrfs_path *p,
		       struct extent_buffer *b, int level, int ins_len,
		       int *write_lock_level)
{
	int ret;
	if ((p->search_for_split || ins_len > 0) && btrfs_header_nritems(b) >=
	    BTRFS_NODEPTRS_PER_BLOCK(root) - 3) {
		int sret;

		if (*write_lock_level < level + 1) {
			*write_lock_level = level + 1;
			kgr_btrfs_release_path(p);
			goto again;
		}

		kgr_btrfs_set_path_blocking(p);
		kgr_reada_for_balance(root, p, level);
		sret = kgr_split_node(trans, root, p, level);
		kgr_btrfs_clear_path_blocking(p, NULL, 0);

		BUG_ON(sret > 0);
		if (sret) {
			ret = sret;
			goto done;
		}
		b = p->nodes[level];
	} else if (ins_len < 0 && btrfs_header_nritems(b) <
		   BTRFS_NODEPTRS_PER_BLOCK(root) / 2) {
		int sret;

		if (*write_lock_level < level + 1) {
			*write_lock_level = level + 1;
			kgr_btrfs_release_path(p);
			goto again;
		}

		kgr_btrfs_set_path_blocking(p);
		kgr_reada_for_balance(root, p, level);
		sret = kgr_balance_level(trans, root, p, level);
		kgr_btrfs_clear_path_blocking(p, NULL, 0);

		if (sret) {
			ret = sret;
			goto done;
		}
		b = p->nodes[level];
		if (!b) {
			kgr_btrfs_release_path(p);
			goto again;
		}
		BUG_ON(btrfs_header_nritems(b) == 1);
	}
	return 0;

again:
	ret = -EAGAIN;
done:
	return ret;
}

static void kgr_reada_for_search(struct btrfs_root *root,
			     struct btrfs_path *path,
			     int level, int slot, u64 objectid)
{
	struct extent_buffer *node;
	struct btrfs_disk_key disk_key;
	u32 nritems;
	u64 search;
	u64 target;
	u64 nread = 0;
	u64 gen;
	int direction = path->reada;
	struct extent_buffer *eb;
	u32 nr;
	u32 blocksize;
	u32 nscan = 0;

	if (level != 1)
		return;

	if (!path->nodes[level])
		return;

	node = path->nodes[level];

	search = btrfs_node_blockptr(node, slot);
	blocksize = btrfs_level_size(root, level - 1);
	eb = kgr_btrfs_find_tree_block(root, search, blocksize);
	if (eb) {
		kgr_free_extent_buffer(eb);
		return;
	}

	target = search;

	nritems = btrfs_header_nritems(node);
	nr = slot;

	while (1) {
		if (direction < 0) {
			if (nr == 0)
				break;
			nr--;
		} else if (direction > 0) {
			nr++;
			if (nr >= nritems)
				break;
		}
		if (path->reada < 0 && objectid) {
			kgr_btrfs_node_key(node, &disk_key, nr);
			if (btrfs_disk_key_objectid(&disk_key) != objectid)
				break;
		}
		search = btrfs_node_blockptr(node, nr);
		if ((search <= target && target - search <= 65536) ||
		    (search > target && search - target <= 65536)) {
			gen = btrfs_node_ptr_generation(node, nr);
			kgr_readahead_tree_block(root, search, blocksize, gen);
			nread += blocksize;
		}
		nscan++;
		if ((nread > 65536 || nscan > 32))
			break;
	}
}

/* is isra, needs to be here */
static int
kgr_read_block_for_search(struct btrfs_trans_handle *trans,
		       struct btrfs_root *root, struct btrfs_path *p,
		       struct extent_buffer **eb_ret, int level, int slot,
		       struct btrfs_key *key, u64 time_seq)
{
	u64 blocknr;
	u64 gen;
	u32 blocksize;
	struct extent_buffer *b = *eb_ret;
	struct extent_buffer *tmp;
	int ret;

	blocknr = btrfs_node_blockptr(b, slot);
	gen = btrfs_node_ptr_generation(b, slot);
	blocksize = btrfs_level_size(root, level - 1);

	tmp = kgr_btrfs_find_tree_block(root, blocknr, blocksize);
	if (tmp) {
		/* first we do an atomic uptodate check */
		if (kgr_btrfs_buffer_uptodate(tmp, gen, 1) > 0) {
			*eb_ret = tmp;
			return 0;
		}

		/* the pages were up to date, but we failed
		 * the generation number check.  Do a full
		 * read for the generation number that is correct.
		 * We must do this without dropping locks so
		 * we can trust our generation number
		 */
		kgr_btrfs_set_path_blocking(p);

		/* now we're allowed to do a blocking uptodate check */
		ret = kgr_btrfs_read_buffer(tmp, gen);
		if (!ret) {
			*eb_ret = tmp;
			return 0;
		}
		kgr_free_extent_buffer(tmp);
		kgr_btrfs_release_path(p);
		return -EIO;
	}

	/*
	 * reduce lock contention at high levels
	 * of the btree by dropping locks before
	 * we read.  Don't release the lock on the current
	 * level because we need to walk this node to figure
	 * out which blocks to read.
	 */
	kgr_btrfs_unlock_up_safe(p, level + 1);
	kgr_btrfs_set_path_blocking(p);

	kgr_free_extent_buffer(tmp);
	if (p->reada)
		kgr_reada_for_search(root, p, level, slot, key->objectid);

	kgr_btrfs_release_path(p);

	ret = -EAGAIN;
	tmp = kgr_read_tree_block(root, blocknr, blocksize, 0);
	if (tmp) {
		/*
		 * If the read above didn't mark this buffer up to date,
		 * it will never end up being up to date.  Set ret to EIO now
		 * and give up so that our caller doesn't loop forever
		 * on our EAGAINs.
		 */
		if (!kgr_btrfs_buffer_uptodate(tmp, 0, 0))
			ret = -EIO;
		kgr_free_extent_buffer(tmp);
	}
	return ret;
}

static inline void kgr_btrfs_tree_unlock_rw(struct extent_buffer *eb, int rw)
{
	if (rw == BTRFS_WRITE_LOCK || rw == BTRFS_WRITE_LOCK_BLOCKING)
		kgr_btrfs_tree_unlock(eb);
	else if (rw == BTRFS_READ_LOCK_BLOCKING)
		kgr_btrfs_tree_read_unlock_blocking(eb);
	else if (rw == BTRFS_READ_LOCK)
		kgr_btrfs_tree_read_unlock(eb);
	else
		BUG();
}

int kgr_btrfs_search_slot(struct btrfs_trans_handle *trans,
	struct btrfs_root *root, struct btrfs_key *key, struct btrfs_path *p,
	int ins_len, int cow)
{
	struct extent_buffer *b;
	int slot;
	int ret;
	int err;
	int level;
	int lowest_unlock = 1;
	int root_lock;
	/* everything at write_lock_level or lower must be write locked */
	int write_lock_level = 0;
	u8 lowest_level = 0;
	int min_write_lock_level;
	int prev_cmp;

	lowest_level = p->lowest_level;
	WARN_ON(lowest_level && ins_len > 0);
	WARN_ON(p->nodes[0] != NULL);
	BUG_ON(!cow && ins_len);

	if (ins_len < 0) {
		lowest_unlock = 2;

		/* when we are removing items, we might have to go up to level
		 * two as we update tree pointers  Make sure we keep write
		 * for those levels as well
		 */
		write_lock_level = 2;
	} else if (ins_len > 0) {
		/*
		 * for inserting items, make sure we have a write lock on
		 * level 1 so we can update keys
		 */
		write_lock_level = 1;
	}

	if (!cow)
		write_lock_level = -1;

	if (cow && (p->keep_locks || p->lowest_level))
		write_lock_level = BTRFS_MAX_LEVEL;

	min_write_lock_level = write_lock_level;

again:
	prev_cmp = -1;
	/*
	 * we try very hard to do read locks on the root
	 */
	root_lock = BTRFS_READ_LOCK;
	level = 0;
	if (p->search_commit_root) {
		/*
		 * the commit roots are read only
		 * so we always do read locks
		 */
		b = root->commit_root;
		extent_buffer_get(b);
		level = btrfs_header_level(b);
		if (!p->skip_locking)
			kgr_btrfs_tree_read_lock(b);
	} else {
		if (p->skip_locking) {
			b = kgr_btrfs_root_node(root);
			level = btrfs_header_level(b);
		} else {
			/* we don't know the level of the root node
			 * until we actually have it read locked
			 */
			b = kgr_btrfs_read_lock_root_node(root);
			level = btrfs_header_level(b);
			if (level <= write_lock_level) {
				/* whoops, must trade for write lock */
				kgr_btrfs_tree_read_unlock(b);
				kgr_free_extent_buffer(b);
				b = kgr_btrfs_lock_root_node(root);
				root_lock = BTRFS_WRITE_LOCK;

				/* the level might have changed, check again */
				level = btrfs_header_level(b);
			}
		}
	}
	p->nodes[level] = b;
	if (!p->skip_locking)
		p->locks[level] = root_lock;

	while (b) {
		level = btrfs_header_level(b);

		/*
		 * setup the path here so we can release it under lock
		 * contention with the cow code
		 */
		if (cow) {
			/*
			 * if we don't really need to cow this block
			 * then we don't want to set the path blocking,
			 * so we test it here
			 */
			if (!kgr_should_cow_block(trans, root, b))
				goto cow_done;

			kgr_btrfs_set_path_blocking(p);

			/*
			 * must have write locks on this node and the
			 * parent
			 */
			if (level > write_lock_level ||
			    (level + 1 > write_lock_level &&
			    level + 1 < BTRFS_MAX_LEVEL &&
			    p->nodes[level + 1])) {
				write_lock_level = level + 1;
				kgr_btrfs_release_path(p);
				goto again;
			}

			err = kgr_btrfs_cow_block(trans, root, b,
					      p->nodes[level + 1],
					      p->slots[level + 1], &b);
			if (err) {
				ret = err;
				goto done;
			}
		}
cow_done:
		p->nodes[level] = b;
		kgr_btrfs_clear_path_blocking(p, NULL, 0);

		/*
		 * we have a lock on b and as long as we aren't changing
		 * the tree, there is no way to for the items in b to change.
		 * It is safe to drop the lock on our parent before we
		 * go through the expensive btree search on b.
		 *
		 * If we're inserting or deleting (ins_len != 0), then we might
		 * be changing slot zero, which may require changing the parent.
		 * So, we can't drop the lock until after we know which slot
		 * we're operating on.
		 */
		if (!ins_len && !p->keep_locks) {
			int u = level + 1;

			if (u < BTRFS_MAX_LEVEL && p->locks[u]) {
				kgr_btrfs_tree_unlock_rw(p->nodes[u], p->locks[u]);
				p->locks[u] = 0;
			}
		}

		ret = kgr_key_search(b, key, level, &prev_cmp, &slot);

		if (level != 0) {
			int dec = 0;
			if (ret && slot > 0) {
				dec = 1;
				slot -= 1;
			}
			p->slots[level] = slot;
			err = kgr_setup_nodes_for_search(trans, root, p, b, level,
					     ins_len, &write_lock_level);
			if (err == -EAGAIN)
				goto again;
			if (err) {
				ret = err;
				goto done;
			}
			b = p->nodes[level];
			slot = p->slots[level];

			/*
			 * slot 0 is special, if we change the key
			 * we have to update the parent pointer
			 * which means we must have a write lock
			 * on the parent
			 */
			if (slot == 0 && ins_len &&
			    write_lock_level < level + 1) {
				write_lock_level = level + 1;
				kgr_btrfs_release_path(p);
				goto again;
			}

			kgr_unlock_up(p, level, lowest_unlock,
				  min_write_lock_level, &write_lock_level);

			if (level == lowest_level) {
				if (dec)
					p->slots[level]++;
				goto done;
			}

			err = kgr_read_block_for_search(trans, root, p,
						    &b, level, slot, key, 0);
			if (err == -EAGAIN)
				goto again;
			if (err) {
				ret = err;
				goto done;
			}

			if (!p->skip_locking) {
				level = btrfs_header_level(b);
				if (level <= write_lock_level) {
					err = kgr_btrfs_try_tree_write_lock(b);
					if (!err) {
						kgr_btrfs_set_path_blocking(p);
						kgr_btrfs_tree_lock(b);
						kgr_btrfs_clear_path_blocking(p, b,
								  BTRFS_WRITE_LOCK);
					}
					p->locks[level] = BTRFS_WRITE_LOCK;
				} else {
					err = kgr_btrfs_try_tree_read_lock(b);
					if (!err) {
						kgr_btrfs_set_path_blocking(p);
						kgr_btrfs_tree_read_lock(b);
						kgr_btrfs_clear_path_blocking(p, b,
								  BTRFS_READ_LOCK);
					}
					p->locks[level] = BTRFS_READ_LOCK;
				}
				p->nodes[level] = b;
			}
		} else {
			p->slots[level] = slot;
			if (ins_len > 0 &&
			    kgr_btrfs_leaf_free_space(root, b) < ins_len) {
				if (write_lock_level < 1) {
					write_lock_level = 1;
					kgr_btrfs_release_path(p);
					goto again;
				}

				kgr_btrfs_set_path_blocking(p);
				err = kgr_split_leaf(trans, root, key,
						 p, ins_len, ret == 0);
				kgr_btrfs_clear_path_blocking(p, NULL, 0);

				BUG_ON(err > 0);
				if (err) {
					ret = err;
					goto done;
				}
			}
			if (!p->search_for_split)
				kgr_unlock_up(p, level, lowest_unlock,
					  min_write_lock_level, &write_lock_level);
			goto done;
		}
	}
	ret = 1;
done:
	/*
	 * we don't really know what they plan on doing with the path
	 * from here on, so for now just mark it as blocking
	 */
	if (!p->leave_spinning)
		kgr_btrfs_set_path_blocking(p);
	/*
	 * Fix for CVE-2014-9710
	 *
	 * old: if (ret < 0)
	 */
	if (ret < 0 && !p->skip_release_on_error)
		kgr_btrfs_release_path(p);
	return ret;
}

static struct btrfs_dir_item *kgr_btrfs_match_dir_item_name(struct btrfs_root *root,
			      struct btrfs_path *path,
			      const char *name, int name_len)
{
	struct btrfs_dir_item *dir_item;
	unsigned long name_ptr;
	u32 total_len;
	u32 cur = 0;
	u32 this_len;
	struct extent_buffer *leaf;

	leaf = path->nodes[0];
	dir_item = btrfs_item_ptr(leaf, path->slots[0], struct btrfs_dir_item);
	if (kgr_verify_dir_item(root, leaf, dir_item))
		return NULL;

	total_len = btrfs_item_size_nr(leaf, path->slots[0]);
	while (cur < total_len) {
		this_len = sizeof(*dir_item) +
			btrfs_dir_name_len(leaf, dir_item) +
			btrfs_dir_data_len(leaf, dir_item);
		name_ptr = (unsigned long)(dir_item + 1);

		if (btrfs_dir_name_len(leaf, dir_item) == name_len &&
		    kgr_memcmp_extent_buffer(leaf, name, name_ptr, name_len) == 0)
			return dir_item;

		cur += this_len;
		dir_item = (struct btrfs_dir_item *)((char *)dir_item +
						     this_len);
	}
	return NULL;
}

/*
 * Fix for CVE-2014-9710
 *
 * Completely changed function
 */
int kgr_do_setxattr(struct btrfs_trans_handle *trans,
	struct inode *inode, const char *name, const void *value, size_t size,
	int flags)
{
	struct btrfs_dir_item *di = NULL;
	struct btrfs_root *root = BTRFS_I(inode)->root;
	struct btrfs_path *path;
	size_t name_len = strlen(name);
	int ret = 0;

	if (name_len + size > BTRFS_MAX_XATTR_SIZE(root))
		return -ENOSPC;

	path = kgr_btrfs_alloc_path();
	if (!path)
		return -ENOMEM;
	path->skip_release_on_error = 1;

	if (!value) {
		di = kgr_btrfs_lookup_xattr(trans, root, path, btrfs_ino(inode),
					name, name_len, -1);
		if (!di && (flags & XATTR_REPLACE))
			ret = -ENODATA;
		else if (di)
			ret = kgr_btrfs_delete_one_dir_name(trans, root, path, di);
		goto out;
	}

	/*
	 * For a replace we can't just do the insert blindly.
	 * Do a lookup first (read-only btrfs_search_slot), and return if xattr
	 * doesn't exist. If it exists, fall down below to the insert/replace
	 * path - we can't race with a concurrent xattr delete, because the VFS
	 * locks the inode's i_mutex before calling setxattr or removexattr.
	 */
	if (flags & XATTR_REPLACE) {
		ASSERT(mutex_is_locked(&inode->i_mutex));
		di = kgr_btrfs_lookup_xattr(NULL, root, path, btrfs_ino(inode),
					name, name_len, 0);
		if (!di) {
			ret = -ENODATA;
			goto out;
		}
		kgr_btrfs_release_path(path);
		di = NULL;
	}

	ret = kgr_btrfs_insert_xattr_item(trans, root, path, btrfs_ino(inode),
				      name, name_len, value, size);
	if (ret == -EOVERFLOW) {
		/*
		 * We have an existing item in a leaf, split_leaf couldn't
		 * expand it. That item might have or not a dir_item that
		 * matches our target xattr, so lets check.
		 */
		ret = 0;
		kgr_btrfs_assert_tree_locked(path->nodes[0]);
		di = kgr_btrfs_match_dir_item_name(root, path, name, name_len);
		if (!di && !(flags & XATTR_REPLACE)) {
			ret = -ENOSPC;
			goto out;
		}
	} else if (ret == -EEXIST) {
		ret = 0;
		di = kgr_btrfs_match_dir_item_name(root, path, name, name_len);
		ASSERT(di); /* logic error */
	} else if (ret) {
		goto out;
	}

	if (di && (flags & XATTR_CREATE)) {
		ret = -EEXIST;
		goto out;
	}

	if (di) {
		/*
		 * We're doing a replace, and it must be atomic, that is, at
		 * any point in time we have either the old or the new xattr
		 * value in the tree. We don't want readers (getxattr and
		 * listxattrs) to miss a value, this is specially important
		 * for ACLs.
		 */
		const int slot = path->slots[0];
		struct extent_buffer *leaf = path->nodes[0];
		const u16 old_data_len = btrfs_dir_data_len(leaf, di);
		const u32 item_size = btrfs_item_size_nr(leaf, slot);
		const u32 data_size = sizeof(*di) + name_len + size;
		struct btrfs_item *item;
		unsigned long data_ptr;
		char *ptr;

		if (size > old_data_len) {
			if (kgr_btrfs_leaf_free_space(root, leaf) <
			    (size - old_data_len)) {
				ret = -ENOSPC;
				goto out;
			}
		}

		if (old_data_len + name_len + sizeof(*di) == item_size) {
			/* No other xattrs packed in the same leaf item. */
			if (size > old_data_len)
				kgr_btrfs_extend_item(root, path,
						  size - old_data_len);
			else if (size < old_data_len)
				kgr_btrfs_truncate_item(root, path, data_size, 1);
		} else {
			/* There are other xattrs packed in the same item. */
			ret = kgr_btrfs_delete_one_dir_name(trans, root, path, di);
			if (ret)
				goto out;
			kgr_btrfs_extend_item(root, path, data_size);
		}

		item = btrfs_item_nr(slot);
		ptr = btrfs_item_ptr(leaf, slot, char);
		ptr += btrfs_item_size(leaf, item) - data_size;
		di = (struct btrfs_dir_item *)ptr;
		btrfs_set_dir_data_len(leaf, di, size);
		data_ptr = ((unsigned long)(di + 1)) + name_len;
		kgr_write_extent_buffer(leaf, value, data_ptr, size);
		kgr_btrfs_mark_buffer_dirty(leaf);
	} else {
		/*
		 * Insert, and we had space for the xattr, so path->slots[0] is
		 * where our xattr dir_item is and btrfs_insert_xattr_item()
		 * filled it.
		 */
	}
out:
	kgr_btrfs_free_path(path);
	return ret;
}

static int kgr_patch_cve_2014_9710_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}
		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_cve_2014_9710_module_notify(struct notifier_block *nb,
                unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, "btrfs"))
		return 0;

	ret = kgr_patch_cve_2014_9710_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_cve_2014_9710_module_nb = {
	.notifier_call = kgr_patch_cve_2014_9710_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_cve_2014_9710_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module("btrfs")) {
		ret = kgr_patch_cve_2014_9710_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_cve_2014_9710_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_cve_2014_9710_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_cve_2014_9710_module_nb);
}
